import inspect

def dualizable(duals=[]):

    def _dualizable(cls):

        class Dual(object):

            class _Trans(object):

                def __init__(self, instance, dual_map):
                    self.instance = instance
                    self.dual_map = dual_map


                def __getattr__(self, name):

                    _name = self.dual_map.get(name, name)
                    return getattr(self.instance, _name)

                    
            def __init__(self, *args, **kwargs):

                self.instance = cls(*args, **kwargs)

                dual_map = dict(duals
                                + [(d[1], d[0]) for d in duals])

                self.trans = Dual._Trans(self.instance, dual_map)


            def __getattr__(self, name):

                if 'dual' == name:
                    return self.trans

                return getattr(self.instance, name)


        _vars = vars(cls)
        for v in _vars.keys():
            x = getattr(cls, v)

            if inspect.isclass(x):
                setattr(Dual, v, _vars[v])

            elif inspect.ismethod(x):
                setattr(Dual, v, _vars[v])

            elif inspect.isfunction(x) and not '__' == v[0:2]:
                setattr(Dual, v, _vars[v])


        return Dual

    return _dualizable



from . import Value

class DistSys:

    def __init__(self, infs):

        self.infs = infs

        clas = [i.proximal for i in infs] + [i.distal for i in infs]

        self._clas = {c.index:c for c in clas}
        self.clas = list(self._clas.values())
        self.indexes = tuple(self._clas.keys())


    def get_cla(self, ix):
        return self._clas.get(ix, None)


    def get_infomorphisms(self, prox_ix, dist_ix):

        if Value.Any == prox_ix:
            return [inf for inf in self.infs
                    if inf.distal.index == dist_ix]

        if Value.Any == dist_ix:
            return [inf for inf in self.infs
                    if inf.proximal.index == prox_ix]

        return [inf for inf in self.infs
                if inf.proximal.index == prox_ix
                and inf.distal.index == dist_ix]



from . import LinAlg as A
from. import Set as S

class Relation:

    def __init__(self, r_dict):

        self.left = set(r_dict.keys())
        self.right = S.Set.union(*(r_dict[x] \
                                   for x in r_dict.keys()))

        self.table_map_left = dict(zip(self.left, \
                                       range(0, len(self.left))))

        self.table_map_right = dict(zip(self.right, \
                                        range(0, len(self.right))))

        self.r_dict = r_dict
        self.table = A.Matrix(\
                              [[1 \
                                if a in r_dict \
                                and b in r_dict[a] \
                                else 0

                                for b in self.right]
                               for a in self.left]
                              )


    def holds(self, a, b):

        ix_a = self.table_map_left[a]
        ix_b = self.table_map_right[b]

        return 1 == self.table[ix_a, ix_b]


    def not_holds(self, a, b):

        ix_a = self.table_map_left[a]
        ix_b = self.table_map_right[b]

        return 0 == self.table[ix_a, ix_b]


    def all_pairs(self):

        return {(a,b) \
         for a in self.left
         for b in self.right
         if self.holds(a,b)}


class EqRelation(Relation):

    def __init__(self, sigma, parts):

        self.parts = parts

        if not S.Set.is_partition(sigma, parts):
            msg = "'EqRelation' must generate from a "\
                  "disjoint partition."
            raise ValueError(msg)

        r_dict = {}
        for p in parts:
            for x in p:
                if not x in r_dict:
                    r_dict[x] = p
        
        Relation.__init__(self, r_dict)



import numpy as np

class Matrix:

    def __init__(self, m):
        self.matrix = np.matrix(m, copy=True)


    def rows(self):
        return self.matrix.shape[0]


    def cols(self):
        return self.matrix.shape[1]


    def __getitem__(self, key):
        return self.matrix.__getitem__(key)


    def __setitem__(self, key, value):
        self.matrix.__setitem__(key, value)


    def __lt__(self, other):

        if isinstance(other, Matrix):
            other_value = other.matrix
        else:
            other_value = other

        return Matrix(self.matrix < other_value)


    def __gt__(self, other):

        if isinstance(other, Matrix):
            other_value = other.matrix
        else:
            other_value = other

        return Matrix(self.matrix > other_value)


    def dot(self, other):
        return Matrix(np.dot(self.matrix, other.matrix))


    def transpose(self):
        return Matrix(self.matrix.transpose())


    def column(self, n):
        return Matrix(self.matrix[:,n])


    def row(self, n):
        return Matrix(self.matrix[n,:])


    def add_column(self, vals=None):

        if not vals:
            row_count = self.matrix.shape[0]
            vals = np.matrix(np.zeros((row_count,1)))

        self.matrix = np.append(self.matrix, vals, axis=1)


    def add_row(self, vals=None):

        if not vals:
            col_count = self.matrix.shape[1]
            vals = np.matrix(np.zeros(col_count))

        self.matrix = np.append(self.matrix, vals, axis=0)


    def zeros(dims):
        return np.zeros(dims)


    def all(self):
        return Matrix(self.matrix.all())


    def logical_or(self, other):
        return Matrix(np.logical_or(self.matrix, other.matrix))



from . import Sequent as S

class JudgeSet:

    def __init__(self, judgements=set()):
        self.judgements = judgements


    def __len__(self):
        return len(self.judgements)


    def __repr__(self):
        return repr(self.judgements)


    def __iter__(self):
        for x in self.judgements:
            yield x


    def __hash__(self):
        individual_hashes = [hash(x) for x in self.judgements]
        return len(self.judgements) * sum(individual_hashes)


    def try_comparison(action):
        try:
            return action()

        except AttributeError:
            raise NotImplementedError


    def __eq__(self, other):

        return\
            self.judgements.issubset(other.judgements)\
            and other.judgements.issubset(self.judgements)


    def __neq__(self, other):
        return not (self == other)

    def __lt__(self, other):

        len_self = len(self.judgements)
        len_other = len(other.judgements)
        
        lt = lambda:\
             len_self < len_other and\
             self.judgements.issubset(other.judgements)

        return JudgeSet.try_comparison(lt)

    def __gt__(self, other):

        len_self = len(self.judgements)
        len_other = len(other.judgements)

        gt = lambda:\
             len_self > len_other and\
             other.judgements.issubset(self.judgements)

        return JudgeSet.try_comparison(gt)


    def __le__(self, other):

        if(self < other):
            return True

        return self == other

    
    def __ge__(self, other):
        return False


    def entails(self, other):
        return S.Sequent(self, other)


    def add(self, item):
        self.judgements.add(item)



import foowise.math.Set as S

import foowise.channels.Cla as C
import foowise.channels.Invariant as I


def get_invariant(cla):

    subset = S.Set.rand_subset(cla.typ)

    while True:
        yield I.Invariant(cla, next(subset))



from . import Value
from . import Cla as C
from . import Invariant as I

from ..math import Set as S

class Channel:

    def colimit(d):

        c_sum = C.Sum(*d.clas)             

        infs = {(inf.proximal.index, inf.distal.index)
                for inf in d.infs}

        sigma_parts = []
        for dist_ix in range(0, len(d.clas)):

            dist_id = c_sum.ident_at(dist_ix)
            dist_infs = d.get_infomorphisms(Value.Any, dist_id)

            for f in dist_infs:
                prox_id = f.proximal.index
                prox_ix = c_sum.index_of(prox_id)

                toks = {x for x in c_sum.tok
                       if x[prox_ix] == f.f_down[x[dist_ix]]}

                sigma_parts.append(toks)

        sigma = S.Set.intersect(*sigma_parts)

        return I.Invariant(c_sum, sigma, dual=True).quotient()



import functools as f
import itertools as it

import random

class Set:

    def are_equal(left, right):
        return left.issubset(right) and right.issubset(left)


    def product(*xs):

        if xs:
            return it.product(*xs)

        return set()


    def union(*xs):
        return f.reduce(lambda y,z: z.union(y), xs, set())


    def intersect(*xs):

        if not xs:
            return set()
        
        return f.reduce(lambda y,z: z.intersection(y), xs, xs[0])


    def is_partition(sigma, parts):

        _parts = list(parts)
        while(_parts):

            p = _parts.pop()
            are_disj = map(lambda x: p.isdisjoint(x), _parts)

            if not all(are_disj):
                return False

        _union_parts = Set.union(*parts)
        if not _union_parts.issubset(sigma) \
           or not sigma.issubset(_union_parts):
            return False

        return True


    def rand_subset(x):

        _x = list(x)
        l_x = len(x)

        def in_range():
            return random.randint(0, l_x-1)

        while True:
            size = random.randint(1, l_x-1)

            rand_ixs = {in_range() for _ in range(0,size)}

            yield {_x[i] for i in rand_ixs}



import uuid as u

class Index:

    def __init__(self, int_value=None):

        self.uuid = u.uuid1()

        if not int_value:
            self.int = self.uuid.int
        else:
            self.int = int_value


    def from_int(int_value):

        if not isinstance(int_value, int):
            msg = "Expected an instance of 'int'."
            raise TypeError(msg)

        return Index(int_value)
        

    def __eq__(self, other):

        if isinstance(other, int):
            return self.int == other

        if hasattr(other, 'int'):
            return self.int == other.int

        if hasattr(other, 'uuid'):
            return self.uuid.int == other.uuid.int

        return NotImplemented


    def __neq__(self, other):
        return not self.__eq__(other)


    def __hash__(self):
        return self.int


    def __str__(self):
        return str(self.int)



from foowise.channels import Cla as C

from foowise.math import Set as S
from foowise.math import Relation as R

class Invariant(R.EqRelation):

    def __init__(self, cla, sigma, dual=False):

        if dual:
            cla = cla.dual

        parts = []
        toks = list(cla.tok)
        tok_reps = {}
        while toks:
            x = toks.pop()

            typs_x = cla.get_types(x, subset=sigma)

            part = { y for y in toks
                     if cla.types_agree(x, y, sigma) }

            for y in part:
                toks.remove(y)

            part.add(x)
            parts.append(part)
            tok_reps[x] = part

        R.EqRelation.__init__(self, cla.tok, parts)

        self.cla = cla
        self.sigma = sigma
        self.isdual = dual

        self.tok_reps = tok_reps


    def __str__(self):
        return str(self.parts)


    def canon_rep(self, x):
        return next(iter({y for y in self.tok_reps.keys()
                              if x in self.tok_reps[y]}))


    def quotient(self):

        c_dict = { }

        for x in self.tok_reps:
            c_dict[x] = self.cla.get_types(x, subset=self.sigma)

        cla = C.Cla(c_dict)

        if self.isdual:
            return cla.dual

        return cla



import functools as f

from foowise.channels import Index as Id
from foowise.channels import InfoPair as I

from foowise.math import Set as S
from foowise.math import LinAlg as Alg
from foowise.math import Dual as D

@D.dualizable(duals=[
    ('tok', 'typ'),
    ('get_tokens', 'get_types'),
    ('tokens_agree', 'types_agree'),
    ('is_valid', 'is_valid_dual'),
    ('is_invalid', 'is_invalid_dual')])
class Cla:

    class ClaTable:

        def __init__(self, validities):

            if not isinstance(validities, dict):
                msg = "Cla.ClaTable() expected 'dict' but got " \
                      "type: " + repr(validities.__class__) + "."
                raise TypeError(msg)

            tok = validities.keys()
            typ = f.reduce(lambda x,y: x.union(y),
                           validities.values(), set())

            typ_count = len(typ)
            typ_to_col = list(zip(typ, range(0, typ_count)))

            self.typ_to_col = dict(typ_to_col)
            col_to_typ = map(lambda p: (p[1], p[0]), typ_to_col)
            self.col_to_typ = dict(col_to_typ)

            tok_count = len(tok)
            tok_to_row = list(zip(tok, range(0, tok_count)))

            self.tok_to_row = dict(tok_to_row)
            row_to_tok = map(lambda p: (p[1], p[0]), tok_to_row)
            self.row_to_tok = dict(row_to_tok)

            self.matrix = Alg.Matrix(
                        [[1
                          if x in validities
                          and alpha in validities[x]
                          else 0

                          for alpha in typ]
                        for x in tok])


            self.next_tok_key = len(self.tok_to_row)
            self.next_typ_key = len(self.typ_to_col)
            

        def __getitem__(self, key):

            (tok, typ) = key

            if tok not in self.tok_to_row:
                return None

            if typ not in self.typ_to_col:
                return None

            tok_ix = self.tok_to_row[tok]
            typ_ix = self.typ_to_col[typ]

            return self.matrix[tok_ix,typ_ix]


        def __setitem__(self, key, value):

            (tok, typ) = key

            tok_ix = self.tok_to_row[tok]
            typ_ix = self.typ_to_col[typ]

            self.matrix[tok_ix][typ_ix] = value


        def __str__(self):

            def _strlen(x):
                return len(str(x))

            toks = [x for x in self.tok_to_row.keys()]
            typs = [t for t in self.typ_to_col.keys()]

            max_col_width = max(map(lambda x: _strlen(x), typs))

            start_col_ix = max(map(lambda x: _strlen(x), toks))

            pad = 1 + max_col_width

            fmt = '\n' + ' ' * (start_col_ix + pad + 1)

            fmt += f.reduce(lambda acc,n:
                            acc + str(n)
                            + ' '*(pad - _strlen(n) + 1), typs, '')

            fmt += '\n\n'

            for x in toks:
                _str = str(x)
                fmt += _str + ' '*(start_col_ix - len(_str) + 4)

                for t in typs:
                    _str = str(self[(x,t)])
                    fmt += _str + ' '* (pad)
                fmt += '\n\n'

            return fmt


        def is_valid(self, tok, typ):

            return 1 == self[(tok,typ)]


        def is_invalid(self, tok, typ):

            return 0 == self[(tok,typ)]


        def row_values(self, tok):

            tok_ix = self.tok_to_row[tok]

            row = self.matrix[tok_ix, :]

            return { self.col_to_typ[ix] : row[0,ix]
                       for ix in self.col_to_typ.keys() }

        def col_values(self, typ):

            typ_ix = self.typ_to_col[typ]

            col = self.matrix[:, typ_ix]
            
            return { self.row_to_tok[ix] : col[ix,0]
                     for ix in self.row_to_tok.keys() }


    def __init__(self, validities, index=None):

        if not isinstance(validities, dict):

            given_type = validities.__class__
            msg = "expected arg type 'dict' but got " \
                  + "'" + repr(given_type) + "'."
            raise TypeError(msg)

        self.index = index

        self.tok = set()
        self.typ = set()

        for x in validities.keys():

            if x:
                self.tok.add(x)

            for t in validities[x]:
                self.typ.add(t)

        self.validities = validities
        self.table = Cla.ClaTable(self.validities)

        if not index:
            index = Id.Index()

        self.index = index

        if NotImplemented == index.__eq__(0):
            msg = "'index' must have an implementation of '__eq__'."
            raise ValueError(msg)


    def is_valid(self, tok, typ):
        return self.table.is_valid(tok, typ)


    def is_valid_dual(self, typ, tok):
        return self.table.is_valid(tok, typ)


    def is_invalid(self, tok, typ):
        return self.table.is_invalid(tok, typ)


    def is_invalid_dual(self, typ, tok):
        return self.table.is_invalid(tok, typ)


    def get_types(self, tok, subset=None):

        if not tok in self.tok:
            return set()

        row = self.table.row_values(tok)

        if not subset:
            return {t for t in row.keys()
                    if 1 == row[t] }

        return { t for t in row.keys()
                 if 1 == row[t] 
                 and t in subset }


    def types_agree(self, x, y, sigma):

        if not x in self.tok:
            raise ValueError(str(x) + " is not a token.")

        if not y in self.tok:
            raise ValueError(str(y) + " is not a token.")

        typ_x = self.get_types(x, subset=sigma)
        typ_y = self.get_types(y, subset=sigma)

        return S.Set.are_equal(typ_x, typ_y)


    def tokens_agree(self, alpha, beta, sigma):

        if not alpha in self.typ:
            raise ValueError(str(alpha) + " is not a type.")

        if not beta in self.typ:
            raise ValueError(str(beta) + " is not a type.")

        tok_x = self.get_tokens(alpha, subset=sigma)
        tok_y = self.get_tokens(beta, subset=sigma)

        return S.Set.are_equal(tok_x, tok_y)


    def get_tokens(self, typ, subset=None):

        if not typ in self.typ:
            return set()

        col = self.table.col_values(typ)

        if not subset:
            return { x for x in col.keys()
                     if 1 == col[x] }

        return { x for x in col.keys()
                 if 1 == col[x] and x in subset }
    

    def empty():
        return Cla({})


class Sum(Cla):

    def __init__(self, *cla):

        indexes = enumerate(map(lambda c: c.index, cla))

        self._ix_to_id = dict(indexes)
        self._id_to_ix = {self._ix_to_id[k]:k for k in
                          self._ix_to_id.keys()}

        tokens = map((lambda c: c.tok), cla)

        tok = S.Set.product(*tokens)

        vals = {x:set() for x in tok}

        for v in vals.keys():

            ix_typs = [
                {(cla[i].index, t)
                 for t in cla[i].get_types(v[i])}
                for i in range(0,len(cla))]

            vals[v] = S.Set.union(*ix_typs)

        Cla.__init__(self, vals)


    def index_of(self, ident):
        return self._id_to_ix.get(ident, None)


    def ident_at(self, index):
        return self._ix_to_id.get(index, None)



class Sequent:

    def __init__(self, antecedent, consequent):
        self.antecedent = antecedent
        self.consequent = consequent


    def __repr__(self):
        return repr(antecdent) + ' |- ' + repr(consequent)


    def __str__(self):
        return str(self)


    def __hash__(self):
        return hash(self.antecedent) * hash(self.consequent)


    def try_comparison(action):
        try:
            return action()
        except:
            raise NotImplemented


    def __eq__(self, other):
        return self.antecedent == other.antecedent\
            and self.consequent == other.consequent


    def __neq__(self, other):
        return not (self == other)


    def __lt__(self, other):

        def lt():
            return\
                self.antecedent < other.antecedent\
                and self.consequent < other.consequent

        return Sequent.try_comparison(lt)


    def __gt__(self, other):

        def gt():
            return self.antecedent > other.antecedent\
            and self.consequent > other.consequent

        return Sequent.try_comparison(gt)


    def __le__(self, other):

        if(self < other):
            return True

        return self == other


    def __ge__(self, other):

        if(self > other):
            return True

        return self == other




import enum as e
import functools as f


class MorphismRangeError(Exception):

    def __init__(self, f_name=None, r_name=None, img=None, target=None):

        if not f_name:
            f_name = 'the given function'
        if not r_name:
            'the expected range'

        message = "The image of '" + f_name \
                  + "' must be a subset of " + r_name + "."

        if img:
            message += "\nImage of " + f_name + ": "
            message += repr(img)

        if target:
            message += "\nImage of " + r_name + ": "
            message += repr(target)

        super(MorphismRangeError, self).__init__(message)


class InfomorphismAxiomError(Exception):

    def __init__(self, violations=[]):
        
        self.violations = violations

        self.message = 'Given mappings violate the fundamental '\
                       'infomorphism axiom.'

        case_count = min(5, len(violations))

        if case_count > 0:
            self.message += 'First ' + str(case_count)\
                       + ' violations: \n'

            self.message += f.reduce(lambda l,v: str(v)+'\n'+l,\
                                     violations, '')

        



from . import Sequent as S
from ..math import LinAlg as m

class Theory:

    def __init__(self, \
                 cla=None,
                 table=None, alphabet=set(), constraints=set()):

        self.alphabet = alphabet
        self.constraints = constraints

        self.cla = cla
        if cla:
            self.table = self.cla.table


    def isconsequent(self, gamma, delta):

        if gamma.entails(delta) in self.constraints:
            return True


        seq = m.Matrix([
            self.to_vector(gamma),
            self.to_vector(delta)
        ])

        cases = m.Matrix.dot(self.table.matrix, seq.transpose())

        not_all_gammas = cases.column(0) < len(gamma)
        at_least_one_delta = cases.column(1) > 0

        return \
            not_all_gammas\
            .logical_or(at_least_one_delta)\
            .all()

    
    def to_vector(self, judges):

        v = m.Matrix.zeros(len(self.table.typ_to_col))

        for x in judges:
            ix = self.table.typ_to_col[x]
            v[ix] = 1

        return v


    def from_classification(cla):

        return Theory(alphabet=cla.typ, cla=cla)



Any = object()

class Value:

    def __init__(self, index, name):

        self.name = name
        self.index = index


    def __eq__(self, other):

        if not (hasattr(other, 'name')
                and hasattr(other, 'index')):

            return NotImplemented

        return (self.index == other.index
                and self.name == other.name)


    def __str__(self):

        index = str(self.index)
        name = str(self.name)

        return "(" + index + ", " + name + ")"



import uuid as u

from . import Cla as C
from . import InfomorphismError as IE
from . import Index as Id

from ..math import Dual as D

@D.dualizable(duals=[
    ('proximal', 'distal'),
    ('f_up', 'f_down')
])
class Infomorphism:

    class InfoPair:

        def __init__(self, x, t, holds):

            self.tok = x
            self.typ = t
            self.holds = holds


        def __eq__(self, other):

            try:
                return (
                    self.tok == other.tok
                    and self.typ == other.typ
                    and self.holds == other.holds)

            except:
                return False


        def __neq__(self, other):
            return not self.__eq__(other)


        def __str__(self):

            if self.holds:
                infix_op = ' |= '
            elif self.holds:
                infix_op = ' |\= '

            return ('< '
                + repr(self.tok) + infix_op  + repr(self.typ)
                + ' >')


        def valid(x, t):        
            return Infomorphism.InfoPair(x, t, True)


        def invalid(x, t):
            return Infomorphism.InfoPair(x, t, False)


    def __init__(self,
                 proximal,
                 distal,
                 f_up,
                 f_down,
                 validate=True):

        self.proximal = proximal
        self.distal = distal

        self.f_down = {x:f_down(x) for x in self.distal.tok}
        self.f_up = {x:f_up(x) for x in self.proximal.typ}

        self.f_down_img = set(self.f_down.values())
        self.f_up_img = set(self.f_up.values())

        if validate:
            self.satisfies_info_axioms()

        self.index = Id.Index()


    def satisfies_info_axioms(self):
            
        if not self.f_down_img.issubset(self.proximal.tok):
            f_name = "f_down"
            r_name = "the proximal classification's token set"

            raise IE.MorphismRangeError(f_name=f_name,
                                        r_name=r_name,
                                        img=self.f_down_img,
                                        target=self.proximal.tok)

        if not self.f_up_img.issubset(self.distal.typ):
            f_name = "f_up"
            r_name = "the distal classification's type set"
            raise IE.MorphismRangeError(f_name=f_name,
                                        r_name=r_name,
                                        img=self.f_up_img,
                                        target=self.distal.typ)

        def failure_case(f_down_x, t, x, f_up_t):
            
            if self.proximal.is_valid(f_down_x, t):
                prox = Infomorphism.InfoPair.valid(f_down_x, t)
            else:
                prox = Infomorphism.InfoPair.invalid(f_down_x, t)

            if self.distal.is_valid(x, f_up_t):
                dist = Infomorphism.InfoPair.valid(x, f_up_t)
            else:
                dist = Infomorphism.InfoPair.invalid(x, f_up_t)

            return (prox, dist)


        violations = []

        toks = [(x, self.f_down[x]) for x in self.distal.tok]
        typs = [(self.f_up[t], t) for t in self.proximal.typ]

        for (x, f_down_x) in toks:
            for (f_up_t, t) in typs:

                if (self.proximal.is_valid(f_down_x, t)
                    != self.distal.is_valid(x, f_up_t)):

                    violations.append(
                        failure_case(f_down_x, t, x, f_up_t))

        if violations:

            raise IE.InfomorphismAxiomError(violations)

        return


    def canon_quot(cla, inv):

        f_up = lambda typ: typ
        f_down = lambda x: inv.canon_rep(x)

        quot = inv.quotient()

        if inv.isdual:
            return Infomorphism(cla, quot, f_down, f_up)

        return Infomorphism(quot, cla, f_up, f_down)



import setuptools

def readme():
    with open('README.rst') as f:
        return f.read()

setuptools.setup(
    name='foowise',
    version='0.9',
    description='An implementation of Barwise-Seligman Channels',
    long_description=readme(),
    
    url='https://github.com/ben-schulz/foowise',
    author='Benjamin Schulz',
    author_email='benjamin.john.schulz@gmail.com',
    license='MIT',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
        ],

    packages=[
        'foowise.channels',
        'foowise.math',
        'foowise.heuristic'
    ],

    install_requires=[
        'numpy'
    ],

    entry_points={
        'console_scripts':['foowise=foowise.command_line:main']
    },

    test_suite='foowise.test.suite',

    include_package_data=True,
    zip_safe=False
)
